#ifndef MMTRACE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "re.h"
#endif

#include "shader_common.h"
#include "nv_hw.h"
#include "NVCtrl.h"

#ifdef DISASM_SHADER_STANDALONE
unsigned int card_family = -1;
#else
static unsigned int inst[4];
#endif

static ShaderFunc vp;
static ShaderFunc fp;
static unsigned int iaddr;

extern int card_type;
extern unsigned int card_family;
extern unsigned int* fb;
extern unsigned      fb_phys;
extern unsigned int  fb_size;

#define CHECK_RANGE(idx, arr) ((idx)<sizeof(_##arr)/sizeof(const char *)) \
	? _##arr[(idx)] : #arr"_OOB"

#define NODS      (1<<0)
#define BRANCH_TR (1<<1)
#define BRANCH_EL (1<<2)
#define BRANCH_EN (1<<3)
#define BRANCH_RE (1<<4)
#define BRANCH_ALL (BRANCH_TR|BRANCH_EL|BRANCH_EN)
#define COUNT_INC (1<<4)
#define COUNT_IND (1<<5)
#define COUNT_NUM (1<<6)
#define COUNT_ALL (COUNT_INC|COUNT_IND|COUNT_NUM)
struct _opcode_info {
	const char *name;
	int numsrc;
	int flags;
};

static struct _opcode_info ops[] = {
	[SOP_ABS  ] = { "ABS", 1, 0 },
	[SOP_ADD  ] = { "ADD", 2, 0 },
	[SOP_ARA  ] = { "ARA", 1, 0 },
	[SOP_ARL  ] = { "ARL", 1, 0 },
	[SOP_ARR  ] = { "ARR", 1, 0 },
    [SOP_BRA  ] = { "BRA", 0, NODS|BRANCH_TR },
    [SOP_BRK  ] = { "BRK", 0, NODS },
    [SOP_CAL  ] = { "CAL", 0, NODS|BRANCH_TR },
	[SOP_CMP  ] = { "CMP", 2, 0 },
	[SOP_COS  ] = { "COS", 1, 0 },
	[SOP_DIV  ] = { "DIV", 2, 0 },
	[SOP_DDX  ] = { "DDX", 1, 0 },
	[SOP_DDY  ] = { "DDY", 1, 0 },
	[SOP_DP2  ] = { "DP2", 2, 0 },
	[SOP_DP2A ] = { "DP2A", 3, 0 },
	[SOP_DP3  ] = { "DP3", 2, 0 },
	[SOP_DP4  ] = { "DP4", 2, 0 },
	[SOP_DPH  ] = { "DPH", 2, 0 },
	[SOP_DST  ] = { "DST", 2, 0 },
	[SOP_EX2  ] = { "EX2", 1, 0 },
	[SOP_EXP  ] = { "EXP", 1, 0 },
	[SOP_FLR  ] = { "FLR", 1, 0 },
	[SOP_FRC  ] = { "FRC", 1, 0 },
	[SOP_IF   ] = { "IF" , 0, NODS|BRANCH_EL|BRANCH_EN },
	[SOP_KIL  ] = { "KIL", 1, 0 },
	[SOP_LG2  ] = { "LG2", 1, 0 },
	[SOP_LIT  ] = { "LIT", 1, 0 },
	[SOP_LOG  ] = { "LOG", 1, 0 },
	[SOP_LOOP ] = { "LOOP", 0, NODS|COUNT_ALL|BRANCH_EN},
	[SOP_LRP  ] = { "LRP", 3, 0 },
	[SOP_MAD  ] = { "MAD", 3, 0 },
	[SOP_MAX  ] = { "MAX", 2, 0 },
	[SOP_MIN  ] = { "MIN", 2, 0 },
	[SOP_MOV  ] = { "MOV", 1, 0 },
	[SOP_MUL  ] = { "MUL", 2, 0 },
	[SOP_NRM  ] = { "NRM", 1, 0 },
	[SOP_PK2H ] = { "PK2H", 1, 0 },
	[SOP_PK2US] = { "PK2US", 1, 0 },
	[SOP_PK4B ] = { "PK4B", 1, 0 },
	[SOP_PK4UB] = { "PK4UB", 1, 0 },
	[SOP_POW  ] = { "POW", 2, 0 },
	[SOP_POPA ] = { "POPA", 0, 0 },
	[SOP_PUSHA] = { "PUSHA", 1, NODS },
	[SOP_RCC  ] = { "RCC", 1, 0 },
	[SOP_RCP  ] = { "RCP", 1, 0 },	
	[SOP_REP  ] = { "REP", 0, NODS|BRANCH_EN|COUNT_NUM },
	[SOP_RET  ] = { "RET", 0, NODS },
	[SOP_RFL  ] = { "RFL", 1, 0 },
	[SOP_RSQ  ] = { "RSQ", 1, 0 },
	[SOP_SCS  ] = { "SCS", 1, 0 },
	[SOP_SEQ  ] = { "SEQ", 2, 0 },
	[SOP_SFL  ] = { "SFL", 2, 0 },
	[SOP_SGE  ] = { "SGE", 2, 0 },
	[SOP_SGT  ] = { "SGT", 2, 0 },
	[SOP_SIN  ] = { "SIN", 1, 0 },
	[SOP_SLE  ] = { "SLE", 2, 0 },
	[SOP_SLT  ] = { "SLT", 2, 0 },
	[SOP_SNE  ] = { "SNE", 2, 0 },
	[SOP_SSG  ] = { "SSG", 1, 0 },
	[SOP_STR  ] = { "STR", 2, 0 },
	[SOP_SUB  ] = { "SUB", 2, 0 },
	[SOP_TEX  ] = { "TEX", 2, 0 },
	[SOP_TXB  ] = { "TXB", 2, 0 },
	[SOP_TXD  ] = { "TXD", 2, 0 },
	[SOP_TXL  ] = { "TXL", 2, 0 },
	[SOP_TXP  ] = { "TXP", 2, 0 },
	[SOP_UP2H ] = { "UP2H", 1, 0 },
	[SOP_UP2US] = { "UP2US", 1, 0 },
	[SOP_UP4B ] = { "UP4B", 1, 0 },
	[SOP_UP4UB] = { "UP4UB", 1, 0 },
	[SOP_X2D  ] = { "X2D", 3, 0 },
	[SOP_XPD  ] = { "XPD", 2, 0 },
	[SOP_NOP  ] = { "NOP", 0, NODS },
};

static const char *_SFR_STRING[] = {
	[SR_POSITION ] = "position",
	[SR_WEIGHT   ] = "weight",
	[SR_NORMAL   ] = "normal",
	[SR_COL0     ] = "color",
	[SR_COL1     ] = "color.secondary",
	[SR_BFC0     ] = "bfc",
	[SR_BFC1     ] = "bfc.secondary",
	[SR_FOGCOORD ] = "fogcoord",
	[SR_POINTSZ  ] = "pointsize",
	[SR_TEXCOORD0] = "texcoord[0]",
	[SR_TEXCOORD1] = "texcoord[1]",
	[SR_TEXCOORD2] = "texcoord[2]",
	[SR_TEXCOORD3] = "texcoord[3]",
	[SR_TEXCOORD4] = "texcoord[4]",
	[SR_TEXCOORD5] = "texcoord[5]",
	[SR_TEXCOORD6] = "texcoord[6]",
	[SR_TEXCOORD7] = "texcoord[7]",
	[SR_FRAGDATA0] = "data[0]",
	[SR_FRAGDATA1] = "data[1]",
	[SR_FRAGDATA2] = "data[2]",
	[SR_FRAGDATA3] = "data[3]",
	[SR_CLIP0    ] = "clip_plane[0]",
	[SR_CLIP1    ] = "clip_plane[1]",
	[SR_CLIP2    ] = "clip_plane[2]",
	[SR_CLIP3    ] = "clip_plane[3]",
	[SR_CLIP4    ] = "clip_plane[4]",
	[SR_CLIP5    ] = "clip_plane[5]",
	[SR_CLIP6    ] = "clip_plane[6]",
	[SR_FACING   ] = "facing",
};
#define SFR_STRING(idx) CHECK_RANGE((idx), SFR_STRING)

static const char *_SWZ_STRING[] = {
	[SSC_X] = "x",
	[SSC_Y] = "y",
	[SSC_Z] = "z",
	[SSC_W] = "w"
};
#define SWZ_STRING(idx) CHECK_RANGE((idx), SWZ_STRING)

static const char *_SP_STRING[]  = { 
	[SP_FLOAT32] = "R",
	[SP_FLOAT16] = "H",
	[SP_FIXED12] = "X",
	[SP_UNKNOWN] = "?"
};
#define SP_STRING(idx) CHECK_RANGE((idx), SP_STRING)

static const char *_SC_STRING[]  = { 
	[SC_FL] = "FL",
	[SC_LT] = "LT",
	[SC_EQ] = "EQ",
	[SC_LE] = "LE",
	[SC_GT] = "GT",
	[SC_NE] = "NE",
	[SC_GE] = "GE",
	[SC_TR] = "TR",
	[SC_UNKNOWN] = "??"
};
#define SC_STRING(idx) CHECK_RANGE((idx), SC_STRING)

static struct _opcode_info *
_get_op_info(int op)
{
	if (op >= (sizeof(ops) / sizeof(struct _opcode_info)))
		return NULL;
	if (ops[op].name == NULL)
		return NULL;
	return &ops[op];
}

const char *disasm_shader_halfop(ShaderFunc *shader, int merged)
{
	static char buf[2][256];
	static char scratch[64];
	struct _opcode_info *opi;
	ShaderOpcode  op;
	ShaderRegFile file;
	ShaderRegSwz  swz;
	int i;

	op  = shader->GetOpcode(shader, merged);
	opi = _get_op_info(op);
	if (!opi)
		return "NO OPINFO!";

#define APPEND(str) strncat(buf[merged], str, sizeof(buf[merged]))
	buf[merged][0] = '\0';

	APPEND(opi->name);
	if (shader->GetPrecision &&
			(!(opi->flags & BRANCH_ALL)) && (!(opi->flags * NODS)) &&
			(op != SOP_NOP))
		APPEND(SP_STRING(shader->GetPrecision(shader)));
	if (shader->SupportsConditional && shader->SupportsConditional(shader)) {
		if (shader->GetConditionUpdate(shader)) {
			APPEND("C");
			snprintf(scratch, sizeof(scratch), "%d",
					shader->GetCondRegID(shader));
			APPEND(scratch);
		}
	}
	if (shader->GetSaturate && shader->GetSaturate(shader))
		APPEND("_SAT");

	if (!(opi->flags & NODS)) {
		int mask = shader->GetDestMask(shader, merged);

		APPEND(" ");
		switch (shader->GetDestFile(shader, merged)) {
		case SRF_ADDRESS:
			snprintf(scratch, sizeof(scratch), "A%d",
					shader->GetDestID(shader, merged));
			APPEND(scratch);
			break;
		case SRF_TEMP:
			snprintf(scratch, sizeof(scratch), "R%d",
					shader->GetDestID(shader, merged));
			APPEND(scratch);
			break;
		case SRF_RESULT:
			APPEND("result.");
			APPEND(SFR_STRING(shader->GetDestID(shader, merged)));
			break;
		default:
			APPEND("BAD_RESULT_FILE");
			break;
		}
		
		if (mask != SMASK_ALL) {
			if (mask) APPEND(".");
			if (mask & SMASK_X) APPEND("x");
			if (mask & SMASK_Y) APPEND("y");
			if (mask & SMASK_Z) APPEND("z");
			if (mask & SMASK_W) APPEND("w");
		}
	}

	if (shader->SupportsConditional && shader->SupportsConditional(shader) &&
			shader->GetConditionTest(shader)) {
		swz = shader->GetCondRegSwizzle(shader);

		APPEND(" (");
		APPEND(SC_STRING(shader->GetCondition(shader)));
		snprintf(scratch, sizeof(scratch), "%d",
				shader->GetCondRegID(shader));
		APPEND(scratch);
		APPEND(".");
		APPEND(SWZ_STRING(swz.x));
		APPEND(SWZ_STRING(swz.y));
		APPEND(SWZ_STRING(swz.z));
		APPEND(SWZ_STRING(swz.w));
		APPEND(")");
	}

	/* looping */
	if (opi->flags & COUNT_ALL) {
		APPEND(" { ");
		if (opi->flags & COUNT_NUM) {
			snprintf(scratch, sizeof(scratch), "%d", shader->GetLoopCount(shader));
			APPEND(scratch);
		}
		if (opi->flags & COUNT_IND) {
			APPEND(", ");
			snprintf(scratch, sizeof(scratch), "%d", shader->GetLoopInitial(shader));
			APPEND(scratch);
		}
		if (opi->flags & COUNT_INC) {
			APPEND(", ");
			snprintf(scratch, sizeof(scratch), "%d", shader->GetLoopIncrement(shader));
			APPEND(scratch);
		}
		APPEND(" }");
	}

	/* branching */
	if (opi->flags & BRANCH_TR) {
		APPEND(" ");
		snprintf(scratch, sizeof(scratch), "%d",
				shader->GetBranch(shader)
				);
		APPEND(scratch);
	}
	if (opi->flags & BRANCH_EL) {
		APPEND(" ELSE ");
		snprintf(scratch, sizeof(scratch), "%d",
				shader->GetBranchElse(shader)
				);
		APPEND(scratch);
	}
	if (opi->flags & BRANCH_EN) {
		APPEND(" END ");
		snprintf(scratch, sizeof(scratch), "%d",
				shader->GetBranchEnd(shader)
				);
		APPEND(scratch);
	}

	if (!(opi->flags & NODS) && opi->numsrc) APPEND(",");
	APPEND(" ");
		
	for (i=0;i<opi->numsrc;i++) {
		if (shader->GetSourceAbs(shader, merged, i))
			APPEND("abs(");
		if (shader->GetSourceNegate(shader, merged, i))
			APPEND("-");

		file = shader->GetSourceFile(shader, merged, i);
		switch (file) {
		case SRF_TEMP:
			snprintf(scratch, sizeof(scratch), "R%d",
					shader->GetSourceID(shader, merged, i));
			APPEND(scratch);
			break;
		case SRF_CONST:
			if (shader->GetSourceIndexed(shader, merged, i)) {
				snprintf(scratch, sizeof(scratch), "c[A%d.%s + 0x%x]",
						shader->GetRelAddressRegID(shader),
						SWZ_STRING(shader->GetRelAddressSwizzle(shader)),
						shader->GetSourceID(shader, merged, i)
						);
			} else {
				float val[4];

				if (shader->GetSourceConstVal) {
					shader->GetSourceConstVal(shader, merged, i, val);
					snprintf(scratch, sizeof(scratch), "{ %.02f, %.02f, %.02f, %.02f }",
							val[0], val[1], val[2], val[3]);
				} else {
					snprintf(scratch, sizeof(scratch), "c[0x%x]",
							shader->GetSourceID(shader, merged, i)
							);
				}
			}
			APPEND(scratch);
			break;
		case SRF_ATTRIB:
			if (shader->GetSourceIndexed(shader, merged, i)) {
				snprintf(scratch, sizeof(scratch), "attrib[A%d.%s + %d]",
						shader->GetRelAddressRegID(shader),
						SWZ_STRING(shader->GetRelAddressSwizzle(shader)),
						shader->GetSourceID(shader, merged, i)
						);
			} else {
				snprintf(scratch, sizeof(scratch), "attrib.%s",
						SFR_STRING(shader->GetSourceID(shader, merged, i))
						);
			}
			APPEND(scratch);
			break;
		case SRF_ADDRESS:
			snprintf(scratch, sizeof(scratch), "A%d",
					shader->GetRelAddressRegID(shader)
					);
			APPEND(scratch);
			break;
		case SRF_TEXUNIT:
			snprintf(scratch, sizeof(scratch), "texture[%d]",
					shader->GetSourceID(shader, merged, i));
			APPEND(scratch);
			break;
		default:
			APPEND("UNKNOWN_FILE");
			break;
		}

		swz = shader->GetSourceSwizzle(shader, merged, i);
		if (file != SRF_ADDRESS && file != SRF_TEXUNIT &&
				(swz.x != SSC_X || swz.y != SSC_Y ||
				 swz.z != SSC_Z || swz.w != SSC_W)) {
			snprintf(scratch, sizeof(scratch),
					".%s%s%s%s",
					SWZ_STRING(swz.x), SWZ_STRING(swz.y),
					SWZ_STRING(swz.z), SWZ_STRING(swz.w)
					);
			APPEND(scratch);
		}

		if (shader->GetSourceAbs(shader, merged, i))
			APPEND(")");
		if (shader->GetSourceScale) {
			int scale = shader->GetSourceScale(shader, merged, i);
			if (scale > 1) {
				snprintf(scratch, sizeof(scratch), "{scaled %dx}", scale);
			}
				
		}
		if (i < (opi->numsrc-1)) APPEND(", ");
	}

	if (shader->IsLastInst(shader)) APPEND(" + END");
	return buf[merged];
}

const char *disasm_shader_inst(ShaderFunc *shader)
{
	static char buf[1024];
	
	if (shader->HasMergedInst(shader)) {
		snprintf(buf, sizeof(buf),
				"\n#\t\tINST %d.0: %s"
				"\n#\t\tINST %d.1: %s",
				iaddr, disasm_shader_halfop(shader, 0),
				iaddr, disasm_shader_halfop(shader, 1)
				);
	} else {
		snprintf(buf, sizeof(buf),
				"\n#\t\tINST %d: %s",
				iaddr, disasm_shader_halfop(shader, 0)
				);
	}

	return buf;
}

void shader_buffer_printf(ShaderFunc *shader, unsigned int *buf)
{
	int is_nop;

	/* Check if starts with 2 NOPs */
	is_nop = (buf[0] & -2) | (buf[1] & -2) | (buf[2] & -2) | (buf[3] & -2);
	if ((is_nop == 0) && ((buf[0] & 1)==0)) {
		is_nop = (buf[4] & -2) | (buf[5] & -2) | (buf[6] & -2) | (buf[7] & -2);
	}
	if (is_nop==0) {
		printf("#\tCan not dump program, starts with 2 NOPs\n");
		return;
	}

	iaddr = 0;
	shader->inst = buf;
	while (1) {
		printf("#\t0x%08x 0x%08x 0x%08x 0x%08x %s\n",
			shader->inst[0], shader->inst[1],
			shader->inst[2], shader->inst[3],
			disasm_shader_inst(shader)
		);

		if (shader->IsLastInst(shader))
			break;

		shader->inst += shader->GetOffsetNext(shader);
		iaddr++;
	}

	/*printf("\n");*/
}

#ifndef DISASM_SHADER_STANDALONE
const char *vshader_set_next_id(unsigned int id, void *data) {
	static char buf[100];

	iaddr = id;
	snprintf(buf, sizeof(buf), "Next upload ID: %d ", id);
	return buf;
}
const char *vshader_set_inst0(unsigned int id, void *data) { inst[0] = id; return ""; }
const char *vshader_set_inst1(unsigned int id, void *data) { inst[1] = id; return ""; }
const char *vshader_set_inst2(unsigned int id, void *data) { inst[2] = id; return ""; }
const char *vshader_set_inst3_print(unsigned int id, void *data)
{
	const char *string;

	inst[3] = id;
	vp.inst = inst;

	string = disasm_shader_inst(&vp);

	iaddr++;
	return string;
}

const char *fshader_from_offset(unsigned int id, void *data) {
	static char rbuf[128];
	unsigned int addr;
	unsigned int *buf;

	if (card_type == NV_CTRL_BUS_TYPE_AGP) {
		addr = find_dma_src_from_dest(id & 0xFFFFFFFe);
		if (addr == -1) {
			snprintf(rbuf, sizeof(rbuf), "#0x%08x (Unable to find DMA record - can't disasm)", id);
			return rbuf;
		}
		buf = gpu_ofs_to_agp(addr);
		if (!buf) {
			snprintf(rbuf, sizeof(rbuf), "#0x%08x (couldn't find local address of program)", id);
			return rbuf;
		}
	}
	else {
		/* Print the shader from the fb for non-agp cards...Only done for NV30, NV40 but it might even work for NV50+.*/
		if ( (card_family == NV40) || (card_family == NV30) || (card_family == G70) ) {
			 
			if (fb != NULL) {
				addr = fb_phys + id;
				buf = &fb[id / 4];
			}
			else {
				snprintf(rbuf, sizeof(rbuf), "#0x%08x (fb memory is not avail - can't disasm)", id);
				return rbuf;
			}
		}
		else {
			snprintf(rbuf, sizeof(rbuf), "#0x%08x (cannot find shader - can't disasm)", id);
			return rbuf;
		}
	}
	printf("#      NV30_TCL_PRIMITIVE_3D_ACTIVE_PROGRAM = 0x%08x\n", id);
	printf("# -- program at 0x%p - 0x%X\n", buf, addr);

	shader_buffer_printf(&fp, buf);
	return "";
}
#endif

void shader_init_funcs(void)
{
	memset(&vp, 0x00, sizeof(vp));
	memset(&fp, 0x00, sizeof(fp));

	switch (card_family) {
	case NV20:	
		NV20VPInitShaderFuncs(&vp);
		break;
	case NV30:
		NV30VPInitShaderFuncs(&vp);
		NV30FPInitShaderFuncs(&fp);
		break;
	case NV40:
	case G70:
		NV40VPInitShaderFuncs(&vp);
		NV40FPInitShaderFuncs(&fp);
		break;
	}
}

#ifdef DISASM_SHADER_STANDALONE
void help_and_abort(void)
{
	printf("Usage: disasm_shader [TYPE] [ARCH] [FILE]\n\n");
	printf("TYPE:\n");
	printf("\tvertex         - FILE is a vertex program\n");
	printf("\tfragment       - FILE is a fragment program\n");
	printf("ARCH:\n");
	printf("\tNV20           - vertex programs only.\n");
	printf("\tNV30/NV40/G70  - vertex/fragment programs\n");
	printf("FILE:\n");
	printf("\tPath of text file containing lines in 0xXXXXXXXX format.\n\n");
	exit (1);
}

int main(int argc, char **argv)
{
	unsigned int buf[4096*4];
	FILE *f;
	int shader_type;

	if (argc < 4)
		help_and_abort();
		
	if (!strncmp(argv[1], "vertex", 6)) shader_type = 0;
	else if (!strncmp(argv[1], "fragment", 8)) shader_type = 1;
	else help_and_abort();
	
	if (!strncasecmp(argv[2], "NV20", 4)) card_family = NV20;
	else if (!strncasecmp(argv[2], "NV30", 4)) card_family = NV30;
	else if (!strncasecmp(argv[2], "NV40", 4)) card_family = NV40;
	else if (!strncasecmp(argv[2], "G70", 4)) card_family = G70;
	else help_and_abort();

	shader_init_funcs();

	f = fopen(argv[3], "r");
	if (!f) help_and_abort();

	int inum = 0;
	unsigned int data;
	while (!feof(f) && inum<sizeof(buf)) {
		if (!fscanf(f, "0x%08x\n", &data)) {
			printf("invalid data in input file\n");
			exit(1);
		}
		buf[inum++] = data;
	}

	if (shader_type)
		shader_buffer_printf(&fp, buf);
	else
		shader_buffer_printf(&vp, buf);

	return 0;
}
#endif


