#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <SDL.h>

#include "re.h"
#include "regl.h"
#include "detection.h"
#include "tests.h"
#include "objects.h"
#include "nv_hw.h"
#include "xvmc_tests.h"
#include "NVCtrl.h"

typedef struct
{
	void (*func)(void);
	const char* name;
} func_store;

#define N_(name)	name, # name

func_store test_funcs[]={
	/*{ N_(test_nehe02) },*/
	{ N_(test_alpha) },
	{ N_(test_blend) },
	{ N_(test_clear) },
	{ N_(test_clip_plane) },
	{ N_(test_color_mask) },
	{ N_(test_color_material) },
	{ N_(test_cull_face) },
	{ N_(test_depth) },
	{ N_(test_display_list) },
	{ N_(test_draw_buffer) },
	{ N_(test_draw_pixels) },
	{ N_(test_enables) },
//	{ N_(test_evaluators) },
	{ N_(test_fb_to_tex) },
	{ N_(test_fog) },
	{ N_(test_fog_hint) }, 
	{ N_(test_front_face) },
	{ N_(test_light_single) },
	{ N_(test_light_two_sides) },
	{ N_(test_lights) },
	{ N_(test_line_hint) }, 
	{ N_(test_line_stipple) },
	{ N_(test_line_width) },
	{ N_(test_logic_op) },
	{ N_(test_matrix) },
	{ N_(test_point_size) },
	{ N_(test_polygon_offset) },
	{ N_(test_polygon_mode) },
	{ N_(test_polygon_stipple) },
	{ N_(test_primitives) },
	{ N_(test_render_mode) },
	{ N_(test_scissor) },
	{ N_(test_shade_model) },
	{ N_(test_single_vertex) },
	{ N_(test_stencil) },
	{ N_(test_tex) },
	{ N_(test_texture_3d) },
	{ N_(test_texture_blend_color) },
	{ N_(test_texture_border) },
	{ N_(test_texture_env) },
	{ N_(test_texture_filter) },
	{ N_(test_texture_format) },
	{ N_(test_texture_mipmap_levels) },
	{ N_(test_texture_size) },
	{ N_(test_texture_spherical_map) },
	{ N_(test_texture_wrap) },
	{ N_(test_viewport) },
	{ N_(test_arb_fragment_program) },
	{ N_(test_arb_imaging) },
	{ N_(test_arb_multisample) },
	{ N_(test_arb_multitexture) },
	{ N_(test_arb_occlusion_query) },
	{ N_(test_arb_point_parameters) },
	{ N_(test_arb_point_sprite) },
	{ N_(test_arb_shadow) },
	{ N_(test_arb_texture_cube_map) },
	{ N_(test_arb_vertex_buffer_object) },
	{ N_(test_arb_vertex_program) },
	{ N_(test_ext_abgr) },
	{ N_(test_ext_bgra) },
	{ N_(test_ext_blend_equation_separate) },
	{ N_(test_ext_blend_func_separate) },
	{ N_(test_ext_blend_logic_op) },
	{ N_(test_ext_fog_coord) },
	/*{ N_(test_ext_framebuffer_blit) },*/
	{ N_(test_ext_framebuffer_object) },
	{ N_(test_ext_shadow_funcs) },
	{ N_(test_ext_stencil_two_side) },
	{ N_(test_ext_texture_filter_anisotropic) },
	{ N_(test_ext_texture_lod_bias) },
	{ N_(test_ext_vertex_weighting) },
	{ N_(test_glsl_vp) },
	{ N_(test_nv_depth_clamp) },
	{ N_(test_nv_fence) },
	{ N_(test_nv_fog_distance) },
	{ N_(test_nv_fragment_program) },
	{ N_(test_nv_fragment_program2) },
	{ N_(test_nv_half_float) },
	{ N_(test_nv_point_sprite) },
	{ N_(test_nv_register_combiners) },
	{ N_(test_nv_register_combiners2) },
	{ N_(test_nv_texgen_emboss) },
	{ N_(test_nv_texture_env_combine4) },
	{ N_(test_nv_texture_expand_normal) },
	{ N_(test_nv_texture_rectangle) },
	{ N_(test_nv_texture_shader) },
	{ N_(test_nv_vertex_program1_1) },
	{ N_(test_nv_vertex_program2) },
	{ N_(test_nv_vertex_program3) },
	{ N_(test_sgis_generate_mipmap) },
	{ N_(test_clip_plane_after_vertex) },
	{ N_(test_interesting_regs) },
	{ N_(test_nv_fragment_program_half) },
	{ N_(test_arb_vertex_buffer_object_indexed) },
//	{ N_(test_exa) }
//	{ N_(test_shadow_tex) }
};

int fifo_channel = -1;
static int fifo_mode_before;
static uint32_t fifo_ctx_table_before[128];

/* Are we running OpenGL mode and tests? */
static int run_opengl = 1;
static int run_fast = 0;

/* Are we running XvMC tests? */
static int run_xvmc = 0;


static void run_opengl_tests();
static void run_xvmc_tests(struct xvmc_state_t *);

void channel_detect_pre()
{
	int i;

	switch (card_family) {
	case G80:
		for (i=0; i<128; i++)
			fifo_ctx_table_before[i] = all_regs[(0x2600+(i*4))/4];
		break;
	default:
		fifo_mode_before = all_regs[0x2504/4];
		break;
	}
}

void channel_detect_post()
{
	int fifo_mode;
	int i;

	switch (card_family) {
	case G80:
		for (i=0; i<128; i++) {
			if (all_regs[(0x2600+(i*4))/4] != fifo_ctx_table_before[i]) {
				if (fifo_channel != -1) {
					printf("Multiple channels enabled since detect_pre()\n");
					goto nv50_fifo_detect_post_error;
				}
				fifo_channel = i;
			}
		}

		if (fifo_channel != -1)
			return;

nv50_fifo_detect_post_error:
		printf("FIFO channel couldn't be detected. PLEASE REPORT\n");
		printf("PFIFO Context table dump (before):\n");
		for (i=0; i<128; i++)
			printf("\tChannel %d: 0x%08x\n", i, fifo_ctx_table_before[i]);
		printf("PFIFO Context table dump (after):\n");
		for (i=0; i<128; i++)
			printf("\tChannel %d: 0x%08x\n", i, all_regs[(0x2600+(i*4))/4]);
		break;
	default:
		fifo_mode = all_regs[0x2504/4] & ~fifo_mode_before;
		if (!fifo_mode) {
			printf("FIFO channel couldn't be detected. PLEASE REPORT"
				   " (0x2504 is %x/%x)\n", fifo_mode_before,
				   all_regs[0x2504/4]);
			return;
		}

		for (fifo_channel=0;fifo_channel<32;fifo_channel++)
			if (fifo_mode & (1<<fifo_channel)) break;
		break;
	}
}

static void cleanup()
{
	destroy_pcilist();
	close_card();
	SDL_Quit();
}

static void print_usage(char *argv0)
{
	printf("Usage: %s [OPTION ...]\n\n", argv0);
	printf("Options:\n"
	       "\t-f, --float\tPrint \"after\" values as IEEE 754 FP32\n"
	       "\t-i, --int\tPrint \"after\" values as INT32\n"
	       "\t-s, --short\tPrint \"after\" values as 2 INT16\n"
	       "\t-c, --char\tPrint \"after\" values as 4 INT8\n"
	       "\t-b, --bit\tPrint \"after\" values as bit pattern\n"
	       "\t-x, --fast\tOnly run startup and default OpenGL tests\n"
	       "\t-d, --database </path/to/renouveau.xml>\trenouveau.xml database\n"
	       "\t-m, --mapping <n> to force usage of mapping n as fifo (BEWARE, NOT FOR KIDS)\n"
	       "\t--xvmc=0|1\tRun XvMC tests [default=%d]\n"
	       "\t--opengl=0|1\tRun OpenGL tests [default=%d]\n",
	       run_xvmc, run_opengl);
}

#define MAXPRINT 40
/* Used by object.c */
data_print_store user_data_print[MAXPRINT];

static void set_user_data_print(int index, int lowbit, int highbit,
				PrintFunc printer, char *data)
{
	if(index > (MAXPRINT-2))
	{
		fprintf(stderr, "ERROR: too many data print formats into "
		        "user_data_print.\n");
	}
	user_data_print[index].lowbit = lowbit;
	user_data_print[index].highbit = highbit;
	user_data_print[index].func = printer;
	user_data_print[index].data = data;
}

static void parse_opts(int argc, char **argv)
{
	int c;
	int i;

	memset(user_data_print, 0, sizeof(data_print_store)*MAXPRINT);
	int cur_pr = 0;

	struct option long_opts[] =
	{
		{ "help", 0, NULL, 'h' },
		{ "fast", 0, NULL, 'x' },
		{ "opengl", 2, NULL, 0 },
		{ "xvmc", 2, NULL, 0 },
		{ "float", 0, NULL, 'f' },
		{ "int", 0, NULL, 'i' },
		{ "short", 0, NULL, 's' },
		{ "char", 0, NULL, 'c' },
		{ "bit", 0, NULL, 'b' },
		{ "database", 1, NULL, 'd' },
		{ "mapping", 1, NULL, 'm' },
		{ 0, 0, NULL, 0 }
	};

	const char *opt_name;
	int opt_index;
	int opt_val;

	while ((c = getopt_long(argc, argv, "xhfiscbdm:", long_opts,
				&opt_index)) != -1)
	{
		switch (c)
		{
		case 0:
			opt_name = long_opts[opt_index].name;
			if (strcmp(opt_name, "mapping")==0) {
				fifo_forced_mapping = atoi(optarg);
			}
			if (strcmp(opt_name, "database")==0) {
				redbxml_filename = optarg;
			}

			if (optarg != NULL &&
			    (!strcmp(optarg, "no") ||
			     !strcmp(optarg, "0") ||
			     !strcmp(optarg, "off"))) {
				opt_val = 0;
			} else {
				opt_val = 1;
			}

			if (!strcmp(opt_name, "opengl")) {
				run_opengl = opt_val;
			} else if (!strcmp(opt_name, "xvmc")) {
				run_xvmc = opt_val;
			}
			break;
		case '?':
			fprintf(stderr, "\nTry `%s -h' for help.\n",
				argv[0]);
			exit(1); 
		case 'h':
			print_usage(argv[0]);
			exit(0);
		case 'f':
			set_user_data_print(cur_pr, 0, 31, print_float, "%f");
			cur_pr++;
			break;
		case 'i':
			set_user_data_print(cur_pr, 0, 31, print_format, "%d");
			break;
		case 's':
			set_user_data_print(cur_pr, 0, 15, print_format, "%d");
			cur_pr++;
			set_user_data_print(cur_pr, 16, 31, print_format, "%d");
			cur_pr++;
			break;
		case 'c':
			for (i = 0; i < 32; i += 8)
			{
				set_user_data_print(cur_pr, i, i + 7,
						    print_format, "%d");
				cur_pr++;
			}
			break;
		case 'b': /* FIXME: this leads to segfault in format_data() */
			fprintf(stderr, "-b is broken\n");
			exit(1);
			for (i = 0; i < 32; i++)
			{
				set_user_data_print(cur_pr, i, i,
						    print_format, "%d");
				cur_pr++;
			}
			break;
		case 'd':
			redbxml_filename = optarg;
			break;
		case 'x':
			run_fast = 1;
			break;
		case 'm':
			fifo_forced_mapping = atoi(optarg);
			break;
		}
	}

	if (cur_pr == 0)
	{
		user_data_print[cur_pr].lowbit = 0;
		user_data_print[cur_pr].highbit = 31;
		user_data_print[cur_pr].func = print_format;
		user_data_print[cur_pr].data = "0x%08x";
		cur_pr++;
	}

}

int main(int argc, char **argv)
{
	SDL_Surface* video_surface;
	struct xvmc_state_t xvmc_state;

	parse_opts(argc, argv);
	atexit(cleanup);
	
	/* stdout will go to file, stderr is left as is */
	redirect_output("stdout");

	if (detect_devices(run_opengl, run_xvmc) != 0)
	{
		fprintf(stderr, "Device detection failed.\n");
		return 1;
	}

	setup_SDL(run_opengl);

	if (run_opengl && !re_load_gl(NULL /*"libGL.so"*/))
	{
		fprintf(stderr, "cannot load GL library\n");
		exit(1);
	}

	/* before we read the maps, we create one more mapping
	 * with all the regs */
	init_card();

	//find_maps();
	//dumpreg_before();
	channel_detect_pre();
	video_surface = SDL_SetVideoMode(512, 512, 0,
	                                 (run_opengl ? SDL_OPENGL : 0) |
	                                 0 /*SDL_NOFRAME*/ /*|SDL_FULLSCREEN*/);
	if (run_xvmc) {
		// Do not try to dump any reg/fifo data for XVMC.
		// The data only exists in a special command buffer.
		// see below in run_xvmc_tests() for more information.
		// Also, trying to dump can cause xids.
		//dumpreg_before();
		//dump_before();
		if (init_xvmc_state(&xvmc_state, 16, 16) < 0) {
			fprintf(stderr, "xvmc: couldn't be initialized.\n");
		} else {
			//dump_after(0);
			//dumpreg_after(0);
		}
	}
	channel_detect_post();
	printf("fifo_channel\t%d\n", fifo_channel);
	//dumpreg_after();

	if (!video_surface)
	{
		printf("unable to set video mode\n");
		exit(0);
	}
	
	// Do XVMC tests first to avoid xids.
	if (run_xvmc) {
		run_xvmc_tests(&xvmc_state);
		/* Rerun init_card after xvmc test, to refresh mappings */
		init_card();
	}

	if (run_opengl) {
		if (!find_maps()) {
			/* Could not find any mapping ? */
		} else if (!fifo_map_found) {
			fprintf(stderr, "FIFO mapping not found, can not dump\n");
		} else {
			run_opengl_tests();
		}
	}

	/* Beware: may crash on some hw */
//	test_interesting_regs();

	/* replace SDL_OPENGL by 0 in SDL_SetVideoMode */
//	test_overlay(video_surface);

	return 0;
}

static void run_opengl_tests()
{
	test_startup();
	printf("# coucou\n");
	test_default();

	if (!run_fast) {
		int i, num_tests = sizeof(test_funcs)/sizeof(test_funcs[0]);	
		for (i=0; i<num_tests; i++) {
			fprintf(stderr, "%d/%d %s"
				"          "
				"          "
				"          "
				"\r", i,num_tests,test_funcs[i].name);
			test_funcs[i].func();

			int error = regl.GetError();
			if (error) {
				fprintf(stderr, "\nLatest OpenGL error: %d\n", error);
			}
		}
		fprintf(stderr,"\n");
	}

	/* set video mode as double buffer, fullscreen, current desktop size */
//	test_swap_buffers();
}

static void run_xvmc_tests(struct xvmc_state_t *xvmc_state)
{
#if 0 // Left in for xorg fifo ref.  This is not really where xvmc is.
	// The reason this worked before for NV4A (and other cards) 
	// is because the xvmc command buffer lies after the xorg fifo
	// on the agp bus.  So, we saw both the xorg fifo and the xvmc command
	// buffer.  
	// XvMC stuff is done elsewhere...
	void *fifo = MAP_FAILED;
	unsigned fifos = all_regs[NV_FIFO_ENABLE/4];
	int fd;

	printf("Fifos state : %x\n", fifos);
	fd = open(device_node_, O_RDWR);
	if ((fifo = mmap(NULL, FIFO_SIZE1, PROT_READ, MAP_SHARED,
			 fd, 0xe0011000-2*0x8000)) == MAP_FAILED)
		perror("mmap");
	close(fd);
#endif
	// XVMC uses a command buffer instead of a fifo. 
	// This command buffer is read by the card and used for doing
	// mpeg2 decoding. 
	// Most likely newer codecs like vc1 and h.264 are implemented 
	// using shaders.  Still, I would have to look at windows to be sure
	// of this.
	int fd;
	int i;
	unsigned int mpeg2_command_buffer_offset;
	unsigned int mpeg2_command_buffer_len;
	unsigned int* mpeg2_command_buffer;
	unsigned int mpeg2_luma_channel_offset;
	unsigned int mpeg2_chroma_channel_offset;
	unsigned int next_command_offset;
	unsigned int channel_size;
	
	// We need the fb phys and card type to see where the command
	// buffer is mapped to.
	extern unsigned fb_phys;
	extern int card_type;
	
	// Command buffer offset and len stored here after context init.
	mpeg2_command_buffer_offset = all_regs[0xB320/4];
	mpeg2_command_buffer_len = all_regs[0xB324/4];
	
	// Information about the first xvmc surface. 
	// These are the luma and chroma channels.  
	// I believe these are tiled vram surfaces.
	mpeg2_luma_channel_offset = all_regs[0xB450/4];
	mpeg2_chroma_channel_offset = all_regs[0xB454/4];
	
	// For non-agp (pci/pcie/integrated), the mpeg2_command_buffer is offset from the fb
	if (card_type != NV_CTRL_BUS_TYPE_AGP)
		mpeg2_command_buffer_offset += fb_phys;
	
	fd = open(device_node_, O_RDWR);
	
	mpeg2_command_buffer = mmap(NULL, mpeg2_command_buffer_len, PROT_READ, MAP_SHARED, fd, mpeg2_command_buffer_offset);
	if (mpeg2_command_buffer == MAP_FAILED) {
		printf("Could not map mpeg2 command buffer");
		mpeg2_command_buffer = NULL;
		return;
	}
	
	close(fd);
	
	// NOTE: We do not need to dump anything because you will not see
	// fifo changes.  The registers might change,I do not believe this
	// is xvmc-releated.
	// Also, we will cause xids if we try to dump.  
	//redirect_output("init_xvmc_window");
	//dump_before();
	//dumpreg_before();
	if (init_xvmc_window(xvmc_state) < 0)
		fprintf(stderr, "xvmc: x11 window creation failed.\n");
	//dumpreg_after();
	//dump_after(0);
	//close_output();
	
	redirect_output("xvmc_command_buffer");
	
	// dump some important xvmc information.
	printf("# MPEG2 Decode Information\n");
	printf("# Command Buffer at 0x%08X, len = 0x%08X\n", 
			mpeg2_command_buffer_offset, mpeg2_command_buffer_len);
	
	// Each channel is allocated from vram (tiled, I believe) using
	// this formula. 
	channel_size = ( (xvmc_state->width + 127) & (~127) ) * 
					 xvmc_state->height;
	printf("# Surface 0 info: Luma = 0x%08X, Chroma = 0x%08X, Size = %d\n", 
			mpeg2_luma_channel_offset, mpeg2_chroma_channel_offset, channel_size);

	test_idct_intra(xvmc_state);
	
	// This is set with the next command offset after the command buffer
	// is filled.  So, the driver sets this register to signal that a 
	// new command is ready to be processed. 
	next_command_offset = all_regs[0xB328/4];
	printf("# Next Command Offset = 0x%08X\n", next_command_offset);
		
	// Just dump the entire command buffer. 
	// Ideally, we would dump after XVMCCreateContext, XXXCreateSurface,
	// XXXRenderSurface.  However, this should suffice for now.
	for (i = 0; i < next_command_offset; i += 4){
		printf("[0x%08X] = 0x%08X\n", i, mpeg2_command_buffer[i/4]);
	}

	printf("# Surface Output\n");
	
	printf("# Luma\n");
	// Dump the macroblock output from both the luma and chroma channels.
	// This is equivalent to dumping a 16x16 image from vram.
	for (i = 0; i < 16; i += 4) {
		printf("[0x%08X] = 0x%08X\n", i, fb[ (mpeg2_luma_channel_offset + i)/4]);
	}
	
	printf("# Chroma\n");
	for (i = 0; i < 16; i += 4) {
		printf("[0x%08X] = 0x%08X\n", i, fb[ (mpeg2_chroma_channel_offset + i)/4]);
	}
	
	munmap(mpeg2_command_buffer, mpeg2_command_buffer_len);
	
	free_xvmc_state(xvmc_state);
#if 0
	if (fifo != MAP_FAILED)
		munmap(fifo, FIFO_SIZE1);
#endif
}
