#ifndef _OBJECTS_H
#define _OBJECTS_H

typedef const char* (*PrintFunc)(unsigned int value, void*data);
const char* print_format(unsigned int value, void*format);
const char* print_float(unsigned int value, void*format);

typedef struct {
	unsigned int id;
	const char* name;
} data_store;

typedef struct {
	int lowbit;
	int highbit;

	PrintFunc func;
	void *data;

	char *name;
} data_print_store;

typedef struct {
	unsigned int offset;
	unsigned int size;
	unsigned int stride;
	const char* name;

	// max 32 bits, so max 32 subfields
	data_print_store data[32];
} object_field_store;

typedef struct {
	unsigned int id;
	const char *name;
	unsigned int card_type;

	unsigned int parent_object;
	unsigned int parent_gpu;

	// You can freely increase this array size, if it's full
	object_field_store fields[0x400];
} object_store;

void object_create(unsigned int parent, unsigned int name, unsigned int type);
unsigned int find_object_type(unsigned int name);
unsigned int get_object_type(unsigned int name);
char *format_load(int object_type, int offset, unsigned int value, int gpu_family);
const char* print_macro(unsigned int value, void*data);
const char* print_enable(unsigned int value, void*data);

//required to be defined elesewhere:
extern data_print_store user_data_print[];

extern const char *record_offset_in(unsigned int, void*);
extern const char *record_offset_out(unsigned int, void*);
extern const char *record_length_in(unsigned int, void*);
extern const char *record_dma_op(unsigned int, void*);

void shader_init_funcs(void);
extern const char *vshader_set_next_id(unsigned int, void*);
extern const char *vshader_set_inst0(unsigned int, void*);
extern const char *vshader_set_inst1(unsigned int, void*);
extern const char *vshader_set_inst2(unsigned int, void*);
extern const char *vshader_set_inst3_print(unsigned int, void*);
extern const char *fshader_from_offset(unsigned int, void *);

extern void clean_dmarecs();

int is_name_dumped(int name);
void set_name_dumped(int name);
#endif
