//#define NV_OLD
#define OUTPUT_MULTIPLE_FILES
#define DONT_DUMP_CHANGED_REGS
#define DONT_DUMP_ABOVE_40000
#define SIMPLE_FIFO_OUTPUT
/*#define USE_XML_FOR_SHADERS 1*/

#include "re.h"
#include "nv_hw.h"
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <string.h>
#include <SDL/SDL_syswm.h>
#include <errno.h>
#ifdef USE_XML_FOR_SHADERS
# include "rexml.h"
#endif

#include "NVCtrl.h"
#include "NVCtrlLib.h"
#include "objects.h"
#include "regl.h"
#include "detection.h"

#define MAPPING_FIFO	1
#define MAPPING_REGS	2
#define MAPPING_NODUMP	4
#define MAPPING_AGP     8


static mapping mappings[100];

static int ignore_regs[]={0x25,0x3b,0x15c,0x15d,0x15e,0x16d};

static long agp_base = 0;
static long agp_size = 0;

unsigned int* fb = NULL;
unsigned      fb_phys;
unsigned int  fb_size;

unsigned int* all_regs = NULL;
unsigned int card_family = 0;
unsigned int pci_device_id = 0;
int          card_type   = -1;
int fifo_map_found = 0;
int fifo_forced_mapping = -1;

const char *redbxml_filename = "renouveau.xml";

typedef struct {
	int type;
	int offset;
} program_t;

static int programs_list_count = 0;
static program_t *programs_list = NULL;

/* Initialize SDL for calling SDL_SetVideoMode with OpenGL.
 * Called from main() and detect_devices().
 */
void setup_SDL(int opengl)
{
	SDL_Init(SDL_INIT_VIDEO);
	if (opengl) {
		//SDL_GL_SetAttribute(SDL_GL_STEREO,1);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,0);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,16);
		SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
		/*SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
		SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 2);*/
	}
}

unsigned int *
gpu_ofs_to_agp(unsigned long ofs)
{
	unsigned long ret = 0;
	int i;

	for (i=0;mappings[i].start != NULL;i++) {
		if (mappings[i].type & MAPPING_AGP) {
			unsigned long addr = ofs + agp_base;
			if( (addr >= mappings[i].phys) &&
			    (addr < (mappings[i].phys + mappings[i].size)) )
			{
				
				ret  = (ofs - (mappings[i].phys - agp_base));
				ret += (unsigned long)mappings[i].start;
				break;
			}
		}
	}

	return (unsigned int *)ret;
}

static int must_ignore(int r)
{
	int i;
	for(i=0;i<sizeof(ignore_regs)/sizeof(int);i++)
		if (ignore_regs[i]==r)
			return 1;
	return 0;
}

static int regcmp(unsigned int* m1,unsigned int* m2,int size)
{
	int i;
	for(i=0;i<size/4;i++)
	{
		if (must_ignore(i))
			continue;
		if (m1[i]!=m2[i])
			return 1;
	}
	return 0;
}

char *bit_pattern(unsigned int val)
{
    static char patbuf[40];
    int count = 32;
    char *p = &patbuf[39];

    *p = '\0';
    p--;

    while (count--) {

        if (val & 1)
            *p = '1';
        else
            *p = '0';

        p--;
        val >>= 1;

        if (!(count % 4)) {
            *p = ' ';
            p--;
        }
    }

    return patbuf;
}

int get_vidram()
{
	int maskedid = pcidev_->device & 0xfff0;
	if( maskedid == 0x01a0 ||   /* nforce */
	    maskedid == 0x01f0 )    /* nforce2 */
	{
		/* on nforces, fail out */
		printf("Note: get_vidram() called on an nForce chip.\n");
		return 0;
	}
	
	if (card_family==NV04)
	{
			int config=all_regs[0x00100000/4]&0x3;
			int amount[4]={32,4,8,16};
			return amount[config];
	} else {
			// won't work on nforces. does anyone have an nforce ?
			return (all_regs[0x0010020c/4]&0xfff00000)>>20;
	}		
}


static int query_bus_type()
{
	int btype = -1;
	/* call NV_CONTROL to find the bus type */
	SDL_SysWMinfo sdli;
	SDL_VERSION(&sdli.version);
	int nvc_base, nvc_err;
	if (SDL_GetWMInfo(&sdli))
	{
		int ok;
		ok = XNVCTRLQueryExtension(sdli.info.x11.display, &nvc_base,
		                           &nvc_err);
		if (ok)
		{
			if (!XNVCTRLQueryAttribute(sdli.info.x11.display, 0,
			      0, NV_CTRL_BUS_TYPE, &btype))
			{
				printf("\nNV-CONTROL query failed.\n");
				return -1;
			}
		} else
			printf("\nNV-CONTROL extension not found.\n");
	} else
		printf("\nFailed to get SDL_SysWMInfo\n");

	return btype;
}

static int memvalid(const void* addr)
{
	int fd=open("/dev/random",O_WRONLY);
	int r=write(fd,addr,1);
	close(fd);
	return (r==1);
}

#ifdef USE_XML_FOR_SHADERS
static void build_programs_list_bitfield(xmlNodePtr node, int obj_type, int obj_offset)
{
	xmlNodePtr arrsz = node->children;

	for(; arrsz; arrsz=arrsz->next) {
		xmlChar *attValue;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "type");
		if (attValue) {
			if (xmlStrcmp(attValue, BAD_CAST "program")==0) {
				/* bitfield with type=program */

				programs_list_count++;
				programs_list = (program_t *) realloc(programs_list, sizeof(program_t) * programs_list_count);
				if (programs_list) {
					programs_list[programs_list_count-1].type = obj_type;
					programs_list[programs_list_count-1].offset = obj_offset;
				}
			}
			xmlFree(attValue);
		}
	}
}

static void build_programs_list_object(xmlNodePtr node, int obj_type)
{
	xmlChar *attValue;
	xmlNodePtr arrsz = node->children;
	int parent_id = 0;

	for(; arrsz; arrsz=arrsz->next) {
		xmlChar *offValue;
		int offset = 0;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		offValue = xmlGetProp(arrsz, BAD_CAST "offset");
		if (offValue) {
			sscanf((char*)offValue, "0x%x", &offset);
			xmlFree(offValue);
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "type");
		if (attValue) {
			if (xmlStrcmp(attValue, BAD_CAST "program")==0) {
				/* type=program */

				programs_list_count++;
				programs_list = (program_t *) realloc(programs_list, sizeof(program_t) * programs_list_count);
				if (programs_list) {
					programs_list[programs_list_count-1].type = obj_type;
					programs_list[programs_list_count-1].offset = offset;
				}

			} else if (xmlStrcmp(attValue, BAD_CAST "bitfield")==0) {
				/* type=bitfield, with a subtype=program ? */

				build_programs_list_bitfield(arrsz, obj_type, offset);
			}
			xmlFree(attValue);
		}
	}

	/* Also add parents */
	attValue = xmlGetProp(node, BAD_CAST "parent");
	if (attValue) {
		if (sscanf((char*)attValue, "0x%x", &parent_id)) {
			xmlNodePtr pr = rexmlGetObjectNode(parent_id);
			build_programs_list_object(pr, obj_type);
		}
		xmlFree(attValue);
	}
}

static void build_programs_list(void)
{
	xmlXPathObjectPtr pathobj;
	xmlNodePtr arrsz;

	pathobj = rexmlSearchNode(BAD_CAST "/renouveau");
	if (!pathobj) {
		return;
	}

	arrsz = pathobj->nodesetval->nodeTab[0]->children;

	for(; arrsz; arrsz=arrsz->next) {
		xmlChar *attValue;
		int id;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "object")!=0) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "id");
		if (sscanf((char*)attValue, "0x%x", &id)) {
			build_programs_list_object(arrsz, id);
		}
		xmlFree(attValue);
	}

	xmlXPathFreeObject(pathobj);
}
#endif /* USE_XML_FOR_SHADERS */

static int is_shader_program(int obj_type, int obj_offset)
{
	int i;

	for (i=0; i<programs_list_count; i++) {
		if (programs_list[i].offset == obj_offset) {
			/* FIXME: renouveau give us only 8 bits for obj_type */
			if ((programs_list[i].type & 0xff) == obj_type) {
				return 1;
			}
		}
	}	

	return 0;
}

static int num_shader_prog = 0;

static void dump_shader_program(int address)
{
	unsigned int addr;
	unsigned int *buf;
	FILE *output;
	char shader_filename[128];
	int is_nop;

	printf("# Dump shader %d from VRAM offset 0x%08x\n", num_shader_prog++, address);

	addr = find_dma_src_from_dest(address & 0xFFFFFFFe);
	if (addr == -1) {
		printf("#  Unable to find DMA record\n");
		return;
	}
	printf("#  DMA source = 0x%08x\n", addr);

	buf = gpu_ofs_to_agp(addr);
	if (!buf) {
		printf("#  Could not find AGP address of program\n");
		return;
	}
	printf("#  AGP source at 0x%08x\n", buf);

	/* Do not dump if starts with 2 NOP */
	is_nop = (buf[0] & -2) | (buf[1] & -2) | (buf[2] & -2) | (buf[3] & -2);
	if ((is_nop == 0) && ((buf[0] & 1)==0)) {
		is_nop = (buf[4] & -2) | (buf[5] & -2) | (buf[6] & -2) | (buf[7] & -2);
	}
	if (is_nop==0) {
		printf("#  Can not dump program, starts with 2 NOPs\n");
		return;
	}

	/* Dump shader to a given file */
	sprintf(shader_filename, "card_10de-%04x_shader_programs.txt",
		pci_device_id & 0xffff);

	output = fopen(shader_filename, (num_shader_prog==1) ? "w" : "a+");
	if (!output) {
		printf("#  Can not create output file\n");
		return;
	}
	
	fprintf(output, "# Shader program %d\n", num_shader_prog-1);
	for (;;) {
		fprintf(output, "%08x %08x %08x %08x\n",
			buf[0], buf[1], buf[2], buf[3]
		);

		if (buf[0] & 1) {
			break;
		}

		/* FIXME: Yep, I know, sometimes 8 */
		buf += 4;
	}

	fclose(output);
}

void init_card()
{
	if (all_regs != NULL)
		return;

	/* first, figure out the physical adress for the regs
	 * These were detected in find_pci_device().
	 */
	unsigned phys_addr, card_id;
	int vid_ram;

	card_id = (pcidev_->vendor << 16) | pcidev_->device;
	pci_device_id = pcidev_->device;
	phys_addr = pcidev_->resource[0].start;
	fb_phys = pcidev_->resource[1].start & 0xFFFFFFF0;
	fb_size = pcidev_->resource[1].len;

	// now we can map it
	int fd;
	fd=open(device_node_, O_RDWR);
	if (fd<0) {
		perror("open");
		exit(1);
	}
	all_regs = mmap(NULL, ALL_REGS_SIZE, PROT_READ|PROT_WRITE,
	                MAP_SHARED, fd,
#ifdef NV_OLD
			0
#else
			phys_addr
#endif
			);
	if (all_regs == MAP_FAILED) {
		close(fd);
		perror("mmap");
		exit(1);
	}

	// now figure out the card family
	card_family = get_card_family(card_id);
	vid_ram = get_vidram();

	printf("pci_id\t0x%08x\n",card_id);
	switch(card_family)
	{
		case NV04:
			printf("# (NV04)\n");
			break;
		case NV10:
			printf("# (NV10)\n");
			break;
		case NV15:
			printf("# (NV15)\n");
			break;
		case NV20:
			printf("# (NV20)\n");
			break;
		case NV30:
			printf("# (NV30)\n");
			break;
		case NV40:
			printf("# (NV40)\n");
			break;
		case G70:
			printf("# (G70)\n");
			break;
		case G80:
			printf("# (G80)\n");
			break;
		default:
			printf("# (unknown)\n");
			break;
	}
#ifdef NV_OLD
	card_type = NV_CTRL_BUS_TYPE_AGP; // sane default :)
#else
	card_type = query_bus_type();
#endif
	switch(card_type)
	{
		case NV_CTRL_BUS_TYPE_AGP:
			printf("bus\tAGP\n");
			break;
		case NV_CTRL_BUS_TYPE_PCI:
			printf("bus\tPCI\n");
			break;
		case NV_CTRL_BUS_TYPE_PCI_EXPRESS:
			printf("bus\tPCI-E\n");
			break;
		case NV_CTRL_BUS_TYPE_INTEGRATED:
			printf("bus\tintegrated\n");
			break;
		default:
			printf("bus\tunknown\n");
			card_type=-1;
			break;
	}
	printf("vram_size\t%d\n",vid_ram);
	printf("BOOT0\t0x%08x\n",all_regs[0]);
	
	if (card_family < G80)
	{
		if (0!=vid_ram && fb_size != vid_ram*(1024*1024))
		{
			printf("# Reported FB size from the PCI config of %dMB is "
				   "different from the actual VRAM size of %dMB\n",
				   fb_size/(1024*1024), vid_ram);
			printf("# Adjusting FB size to match VRAM\n");
			fb_size = vid_ram * 1024 * 1024;
		}
		if (fb_size > 256<<20) {
			printf("# Can only access 256MB of vram\n");
			fb_size = 256<<20;
		}
		printf("fb_size\t%d\n",fb_size>>20);
		
		fb = mmap(NULL, fb_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, fb_phys);
		if (fb == MAP_FAILED) {
			printf("Failed to map video memory!\n");
			fb = NULL;
		}
	}
	close(fd);

	// try to find the address/size of the AGP aperture
	if (card_type == NV_CTRL_BUS_TYPE_AGP)
	{
		pci_resource res;
		if (find_agp_aperture(&res) == 0)
		{
			agp_base = res.start & ~8; // wtf is this?
			agp_size = res.len;
			printf("agp_aperture_base\t0x%lx\n", agp_base);
			printf("agp_aperture_size\t%d\n", (int) (agp_size>>20));
		} else
		{
			agp_base = agp_size = 0;
			printf("# Couldn't determine address of AGP aperture\n");
		}
	}

	shader_init_funcs();
	// figure out which fifos are in use
	/*
	unsigned fifos=all_regs[NV_FIFO_ENABLE/4];
	printf("Fifos state : %x\n",fifos);
	{
		int fd;
		printf("mapping\n");
		fd=open(device_node_, O_RDWR);
		if (mmap(NULL, FIFO_SIZE1, PROT_READ, MAP_SHARED, fd,
		         0xe0011000-2*0x8000) == MAP_FAILED) {
			perror("mmap");
			exit(1);
		}
		close(fd);
	}
	*/

#ifdef SIMPLE_FIFO_OUTPUT
	printf("simple_fifo_output\t1\n");
#else
	printf("simple_fifo_output\t0\n");
#endif

#ifdef USE_XML_FOR_SHADERS
	/* Now load renouveau.xml */
	if (rexmlInit(redbxml_filename)==0) {
		build_programs_list();

		rexmlShutdown();
/*
		int i;
		for (i=0; i<programs_list_count; i++) {
			printf("# Obj 0x%08x offset 0x%08x\n", programs_list[i].type, programs_list[i].offset);
		}
*/
	} else {
		printf("# Could not load renouveau.xml from %s\n", redbxml_filename);
	}

	xmlCleanupParser();
	xmlMemoryDump();
#endif /* USE_XML_FOR_SHADERS */
}

void close_card(void)
{
	if (programs_list) {
		free(programs_list);
		programs_list = NULL;
	}
}

int find_maps(void)
{
	char s[1024];
	int scratch_found=0;
	int mapnum=0;

	FILE *output = fopen("card_mappings.txt", "w");
	if (!output) {
		ERR_PRINT("Can not open card mappings output file\n");
		return 0;
	}

	FILE *f = fopen("/proc/self/maps", "r");
	if (!f) {
		ERR_PRINT("Can not open /proc/self/maps for reading\n");
		return 0;
	}

	while(fgets(s,1024,f))
	{
		if(parse_mapping(s, &mappings[mapnum]))
		{
			ERR_PRINT("Failed to parse line: %s\n", s);
			continue;
		}

		if(!is_wanted_mapping(&mappings[mapnum]))
		{
			/*ERR_PRINT("Discarding line: %s\n", s);*/
			destroy_mapping(&mappings[mapnum]);
			continue;
		}
		
		mappings[mapnum].before = malloc(mappings[mapnum].size);
		mappings[mapnum].after = malloc(mappings[mapnum].size);

		fprintf(output,"map %d\tat %p size 0x%lx\t(physical 0x%lx)\t",
		       mapnum, mappings[mapnum].start, mappings[mapnum].size,
		       mappings[mapnum].phys);
				
		if ( 
		     (mappings[mapnum].size == FIFO_SIZE1) ||
		     (mappings[mapnum].size == FIFO_SIZE2) ||
		     (mappings[mapnum].size == FIFO_SIZE3) ||
		     (mappings[mapnum].size == FIFO_SIZE4) ||
		     (mappings[mapnum].size == FIFO_SIZE6) ||
		     ( (mappings[mapnum].size == FIFO_SIZE5) && scratch_found ) 
		     )
		{
			// the fifo
			mappings[mapnum].type |= MAPPING_FIFO;
			fifo_map_found = 1;
			fprintf(output,"=> fifo\n");
		} else if ( (mappings[mapnum].phys >= agp_base) &&
		            (mappings[mapnum].phys < (agp_base + agp_size)) )
		{
			mappings[mapnum].type |= MAPPING_AGP;
			if (mappings[mapnum].size == SCRATCH_SIZE)
			{
				scratch_found=1;
				fprintf(output,"=> AGP scratch\n");
			}
			else
				fprintf(output,"=> AGP\n");
		} else if ( (mappings[mapnum].size >= 640*480*2) &&
		          (mappings[mapnum].size != REGS_SIZE) &&
		          (mappings[mapnum].size != SCRATCH_SIZE) )
		{
			// the screen or something else we don't want to dump
			mappings[mapnum].type |= MAPPING_NODUMP;
			fprintf(output,"=> no dump\n");
		} else if ( (card_type == NV_CTRL_BUS_TYPE_PCI_EXPRESS) &&
		            ((mappings[mapnum].size == REGS_SIZE2)
			     ||(mappings[mapnum].size == REGS_SIZE3))
			     )
		{
			// crashes when accessed on pci-e
			mappings[mapnum].type |= MAPPING_NODUMP;
			fprintf(output,"=> registers (no read, PCI-E)\n");
		} else if ((mappings[mapnum].size == SCRATCH_SIZE)&&(!scratch_found))
		{
			mappings[mapnum].type |= MAPPING_REGS;
			scratch_found=1;
			fprintf(output,"=> PCI/PCIE scratch\n");
		} else
		{	
			// probably some regs
			mappings[mapnum].type |= MAPPING_REGS;
			fprintf(output,"=> registers\n");
		}
		
		mapnum++;

		// insane quadro hack
		if (mappings[mapnum-1].size == FIFO_SIZE_QUADRO2)
		{
#define QUADRO_FIFO_OFFSET 0xbf400
			mappings[mapnum].start = mappings[mapnum-1].start + QUADRO_FIFO_OFFSET;
			mappings[mapnum].phys = mappings[mapnum-1].phys + QUADRO_FIFO_OFFSET;
			mappings[mapnum].size = FIFO_SIZE2;
			mappings[mapnum].before = (unsigned int*)malloc(mappings[mapnum].size);
			mappings[mapnum].after = (unsigned int*)malloc(mappings[mapnum].size);
			mappings[mapnum].type = MAPPING_FIFO;
			fifo_map_found = 1;
			fprintf(output,"map %d\tfrom 0x%lx to 0x%lx size 0x%lx"
			       " (physical 0x%lx)\t", mapnum,
			       (unsigned long)mappings[mapnum].start,
			       (unsigned long)(mappings[mapnum].start
			                     + mappings[mapnum].size),
			        mappings[mapnum].size, mappings[mapnum].phys);
			fprintf(output,"=> quadro fifo\n");
			mapnum++;
		}
	}
	mappings[mapnum].start=NULL;

	if (fifo_forced_mapping>=0) {
		fprintf(output, "# Warning: forcing fifo type on mapping %d\n", fifo_forced_mapping);
		mappings[fifo_forced_mapping].type = MAPPING_FIFO;
		fifo_map_found = 1;
	}

	clean_dmarecs();

	fclose(f);
	fclose(output);
	return 1;
}



void print_fifo (int n, int before, int after)
{
#ifdef SIMPLE_FIFO_OUTPUT
    printf("%08x", after);
#else
    printf("%x   0x%.8x   0x%.8x   ", n, before, after);
#endif
}

#define ALLREG_START 0x2000
#define ALLREG_SIZE 0x1000
static int32_t allregmapping_before[ALLREG_SIZE/4];
static int32_t allregmapping_after[ALLREG_SIZE/4];

void dumpreg_before()
{
	unsigned char *allreg_src = (unsigned char *)all_regs;
	memcpy(allregmapping_before,&allreg_src[ALLREG_START],ALLREG_SIZE);
}

void dumpreg_after()
{
	unsigned char *allreg_src = (unsigned char *)all_regs;
	memcpy(allregmapping_after,&allreg_src[ALLREG_START],ALLREG_SIZE);

	int j;
	for(j=0;j<ALLREG_SIZE/4;j++)
	{
		if (allregmapping_before[j]!=allregmapping_after[j])
		{
			printf("Changed reg 0x%.8x from 0x%.8x to 0x%.8x\n",
			       ALLREG_START+(j<<2), allregmapping_before[j],
			       allregmapping_after[j]);
		}
	}
}

void dump_before()
{
	int i=0;

	// dump each mapping to its buffer
	while(mappings[i].start)
	{
		if (!memvalid(mappings[i].start))
			mappings[i].type|=MAPPING_NODUMP;
		if (!(mappings[i].type&MAPPING_NODUMP))
			memcpy(mappings[i].before,mappings[i].start,mappings[i].size);
		i++;
	}
}


typedef struct {
	unsigned int name;
	unsigned int type;
} fifo_descriptor;

static fifo_descriptor fifo_channels[8];

static void set_subchannel(int channel, unsigned int name) {
#ifndef SIMPLE_FIFO_OUTPUT
	if (!is_name_dumped(name)) {
		printf("# Mapped subchannel %d to %08x\n", channel, name);
	}
#endif
	fifo_channels[channel].name = name;
	fifo_channels[channel].type = find_object_type(name);
	set_name_dumped(name);
}

const char *dma_opcode_name[4] = {
	"METHOD",
	"JUMP",
	"NONINC_METHOD",
	"CALL",
};

static void dump_changed_regs(mapping *map)
{
	int j;
	if( (map->type & MAPPING_REGS) ||
	    (map->type & MAPPING_AGP) )
	{
		// compare the regs, find the changes
		for(j=0; j<map->size/4; j++)
		{
			if (must_ignore(j))
				continue;
			if (map->before[j] == map->after[j])
				continue;
			printf("# Changed reg 0x%.8x from 0x%.8x to 0x%.8x\n",
			       j, map->before[j], map->after[j]);
		}
	}
}

static void dump_fifo(mapping *map, int dump_startup)
{
	int j;
	unsigned int size, channel, offset, opcode;
	int changed_from = 0x7FFFFFFF, changed_to = -1;
	int upbound;

	if( !(map->type & MAPPING_FIFO) )
		return;

	if (dump_startup)
		changed_from = 0;

#ifdef NV_OLD
	if (map->size == FIFO_SIZE4)
		upbound = 328000;
		/* yet another hack to avoid huge dumps with
		 * 97xx drivers... */
	else
		upbound = map->size;
#else
	upbound = map->size;
	if (upbound > 0x100000)
		upbound = 0x100000;
#endif

	// compare the fifos, find the changes
	for(j=0; j<upbound/4-1; j++)
	{
		if (map->after[j] != map->before[j])
		{
			if (j<changed_from)
				changed_from=j;
			if (j>changed_to)
				changed_to=j;
		}
	}
	changed_to++;
	if (changed_from>changed_to)
	{
		printf("# No fifo changes\n");
		return;
	}

	printf("# Changed from %d to %d   (%d words   %d bytes)\n",
	       changed_from, changed_to, (changed_to-changed_from),
	       (changed_to-changed_from)*4);

	j = 0;
	while (j < changed_from)
	{
		size = (map->after[j] & 0x1ffc0000) >> 18;
		channel = (map->after[j] & 0x0000e000) >> 13;
		offset = map->after[j] & 0x1FFC;
		opcode = map->after[j] >> 29;
		if (offset == 0 && size > 0)
			set_subchannel(channel, map->after[j+1]);

		j += size + 1;
	}

	int nops = 0;
	size = 0;
	for(j=changed_from; j<changed_to || size; j++)
	{
		if (size)
		{
			if (offset == 0) {
				set_subchannel(channel, map->after[j]);
			}

			char *str = format_load(fifo_channels[channel].type,
			                        offset, map->after[j], card_family);
			print_fifo(j, map->before[j], map->after[j]);
#ifndef SIMPLE_FIFO_OUTPUT
			if (offset != 0) {
				printf(str);
			}
#endif
			printf("\n");

#ifdef USE_XML_FOR_SHADERS
			/* Dump shader program ? */
			if (is_shader_program(fifo_channels[channel].type, offset)) {
				dump_shader_program(map->after[j]);
			}
#endif /* USE_XML_FOR_SHADERS */

			size--;
			if (opcode != 2 /* NONINC_METHOD */)
				offset += 4;
			continue;
		}
		
#ifndef SIMPLE_FIFO_OUTPUT
		if (map->after[j] == 0x0) {
			nops ++;
			continue;
		}
						
		if (nops != 0) {
			printf("#\t\t%d NOPs skipped\n", nops);
			nops = 0;
		}
#endif

#ifdef DONT_DUMP_ABOVE_40000
		if (j >= 0x40000) {
			continue;
		}
#endif

		print_fifo(j, map->before[j], map->after[j]);

		size = (map->after[j] & 0x1ffc0000) >> 18;
		channel = (map->after[j] & 0x0000e000) >> 13;
		offset = map->after[j] & 0x1FFC;
		opcode = map->after[j] >> 29;
#ifdef SIMPLE_FIFO_OUTPUT
		printf("\n");
#else
		printf("          {size: 0x%x   channel: 0x%x   "
		       "obj: %08x opcode: %s offset: %04x}\n", size,
		       channel, fifo_channels[channel].name,
		       (opcode<4) ? dma_opcode_name[opcode] : "UNKNOWN",
		       offset);
#endif
	}
}

void dump_after(int dump_startup)
{
	int i;
	
	if (regl.Flush)
		regl.Flush();

	// dump each mapping to its buffer
	i=0;
	while(mappings[i].start)
	{
		if (!memvalid(mappings[i].start))
			mappings[i].type|=MAPPING_NODUMP;
		if (!(mappings[i].type & MAPPING_NODUMP))
			memcpy(mappings[i].after, mappings[i].start,
			       mappings[i].size);
		i++;
	}
	i=0;
	while(mappings[i].start)
	{
		if( mappings[i].type & MAPPING_NODUMP )
		{
			i++;
			continue;
		}

		if( regcmp(mappings[i].after, mappings[i].before,
		           mappings[i].size) == 0 )
		{
			i++;
			continue;
		}

		printf("# ==========================\n");
		printf("# Mapping %d ",i);
		if (mappings[i].type&MAPPING_REGS)
			printf("(regs)");
		if (mappings[i].type&MAPPING_FIFO)
			printf("(fifo)");
		if (mappings[i].type&MAPPING_NODUMP)
			printf("(screen)");

		printf("\n");

#ifndef DONT_DUMP_CHANGED_REGS
		dump_changed_regs(&mappings[i]);
#endif
		dump_fifo(&mappings[i], dump_startup);

		i++;
		printf("\n");
	}	
}

static int _old_fd = -1;
static fpos_t _old_fpos;
void redirect_output(const char* name)
{
#ifdef OUTPUT_MULTIPLE_FILES
	char scratch[1024];

	fflush(stdout);
	fgetpos(stdout, &_old_fpos);
	
	/* preserve the old output target */
	_old_fd = dup(fileno(stdout));
	
	if(pcidev_)
	{
		snprintf(scratch, 1023, "card_%04x-%04x_%s.txt",
		         pcidev_->vendor, pcidev_->device, name);
	} else
	{
		snprintf(scratch, 1023, "card_%s.txt", name);
	}
	
	if(!freopen(scratch, "w", stdout))
	{
		ERR_PRINT("Opening '%s' failed: %s\n",
			scratch, strerror(errno));
	}
#endif
}

/* NOTE: This can be called only once after a redirect_output(),
 * they do not really nest.
 */
void close_output()
{
	fflush(stdout);
#ifdef OUTPUT_MULTIPLE_FILES
	if(-1 == _old_fd)
	{
		ERR_PRINT("no previous stream to return output to, "
		          "continuing with the current output.\n");
		return;
	}

	/* return the previous output file */
	if(dup2(_old_fd, fileno(stdout)) == -1)
		ERR_PRINT("Redirecting to previous file failed: %s\n",
			strerror(errno));
	else
		close(_old_fd);
	_old_fd = -1;
	clearerr(stdout);
	fsetpos(stdout, &_old_fpos);
#endif
}

void let_it_snow()
{
	int i;
	int s;
	i=0;
	s=0;
	while(mappings[i].start)
	{
		if (mappings[i].type&MAPPING_NODUMP)
			s=i;
		i++;
	}
	printf("screen is %d\n",s);
	for(i=0;i<10000000;i++)
	{
		int addr=(rand()%mappings[s].size)&(~0x03);
		*((char*)mappings[s].start+addr)=255;// blue
		*((char*)mappings[s].start+addr+1)=255;// green
		*((char*)mappings[s].start+addr+2)=255;// red
	}
}


